@@ -23,7 +23,8 @@ def attach_profile_lua_scripts(redis_client,
                 created_postfix=PROFILE_VALUE_CREATED_POSTFIX,
                 expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX))
 
-    if not hasattr(redis_client, 'update_profile'):
+    if not hasattr(redis_client, 'update_profile') and \
+            not any(x is None for x in (user_ttl, user_data_ttl_margin)):
 
         redis_client.update_profile = redis_client.register_script(
             LuaScripts.update_profile(
@@ -37,7 +38,8 @@ def attach_profile_lua_scripts(redis_client,
                 expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX,
                 new_user_detection_ttl=new_user_detection_ttl))
 
-    if not hasattr(redis_client, 'get_all_user_pointers'):
+    if not hasattr(redis_client, 'get_all_user_pointers') and \
+            not any(x is None for x in (user_ttl, user_data_ttl_margin)):
 
         redis_client.get_all_user_pointers = redis_client.register_script(
             LuaScripts.get_all_user_pointers(
@@ -45,7 +47,8 @@ def attach_profile_lua_scripts(redis_client,
                 user_ttl_margin=user_data_ttl_margin
             ))
 
-    if not hasattr(redis_client, 'set_profile_if_new_user'):
+    if not hasattr(redis_client, 'set_profile_if_new_user') and \
+            not any(x is None for x in (user_ttl, user_data_ttl_margin, new_user_detection_ttl)):
 
         redis_client.set_profile_if_new_user = redis_client.register_script(
             LuaScripts.set_profile_if_new_user(
@@ -76,7 +79,8 @@ def attach_profile_lua_scripts(redis_client,
                 expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX
             ))
 
-    if not hasattr(redis_client, 'switch_profiles'):
+    if not hasattr(redis_client, 'switch_profiles') and \
+            not any(x is None for x in (user_ttl, user_data_ttl_margin)):
 
         redis_client.switch_profiles = redis_client.register_script(
             LuaScripts.switch_profiles(
@@ -110,10 +114,10 @@ def getProfileClass(
         django_count_func,
         profile_ttl,
         profile_ttl_margin,
-        user_ttl,
-        user_data_ttl_margin,
-        clean_queue_ttl,
-        new_user_detection_ttl):
+        user_ttl=None,
+        user_data_ttl_margin=None,
+        clean_queue_ttl=None,
+        new_user_detection_ttl=None):
 
     cls = Profile
     cls.domain = domain
