@@ -1,5 +1,7 @@
 import datetime
 import pytz
+import time
+import random
 import json
 import uuid
 import functools
@@ -7,10 +9,14 @@ import sys
 import inspect
 
 from python_grains.profiles import profile_properties
+from python_grains.profiles import exceptions
+from python_grains.locks import Lock
 
 DEFAULT_DOMAIN = None
 DEFAULT_REDIS_CLIENT = None
 DEFAULT_DJANGO_PROFILE_MODEL = None
+DEFAULT_DJANGO_COUNT_FUNC = None
+DEFAULT_DJANGO_POINTER_MODEL = None
 DEFAULT_PROFILE_TTL = 12 * 60 * 60
 DEFAULT_PROFILE_TTL_MARGIN = 2 * 60 * 60
 DEFAULT_USER_TTL = 6 * 24 * 60 * 60
@@ -42,6 +48,8 @@ class Profile(object):
     fast_clean_threshold = DEFAULT_FAST_CLEAN_THRESHOLD
     redis_client = DEFAULT_REDIS_CLIENT
     django_profile_model = DEFAULT_DJANGO_PROFILE_MODEL
+    django_pointer_model = DEFAULT_DJANGO_POINTER_MODEL
+    django_count_func = DEFAULT_DJANGO_COUNT_FUNC
     profile_ttl = DEFAULT_PROFILE_TTL
     profile_ttl_margin = DEFAULT_PROFILE_TTL_MARGIN
     user_ttl = DEFAULT_USER_TTL
@@ -49,7 +57,8 @@ class Profile(object):
     clean_queue_ttl = DEFAULT_CLEAN_QUEUE_TTL
 
     def __init__(self,
-                 user_id,
+                 key=None,
+                 user_id=None,
                  profile_id=None,
                  session_id=None,
                  time=None,
@@ -64,6 +73,7 @@ class Profile(object):
 
         self.raw_input = raw_input
         self.user_id = user_id
+        self.key = key
         self.session_id = session_id
         self.updated = updated or time or now
         self.created = created or time or now
@@ -229,13 +239,114 @@ class Profile(object):
 
     @classmethod
     def db_lookup_failed_before(cls, user_id):
+
         return cls.redis_client.get(cls._user_not_found_in_db_key(user_id=user_id)) == b'1'
 
     @classmethod
     def set_db_lookup_fail(cls, user_id):
+
         cls.redis_client.set(cls._user_not_found_in_db_key(user_id=user_id), value='1',
                                         nx=True, ex=cls.db_lookup_fail_check_ttl)
 
+    def store_in_db(self, data=None):
+
+        max_tries = 3
+
+        if self.key is None:
+            self.key = self.new_profile_key
+
+        if data is None:
+            datastring = json.dumps(self.flat_data)
+        else:
+            datastring = json.dumps(data)
+
+        for i in range(max_tries):
+            try:
+                self.django_profile_model.objects.update_or_create(
+                    profile_id=self.profile_id_from_key(self.key),
+                    user_id=self.user_id,
+                    defaults={'data': datastring})
+                return
+            except Exception as e:
+                if 'OperationalError' in e.__class__.__name__:
+                    error = e
+                    time.sleep(1 + random.random() / 10)
+            raise error
+
+    def delete_in_db(self):
+
+        try:
+            self.django_profile_model.objects.get(profile_id=self.profile_id).delete()
+        except self.django_profile_model.DoesNotExist:
+            pass
+
+    def delete_in_cache(self):
+
+        self.redis_client.delete_in_cache(
+            keys=[
+                self.key,
+                self.all_profile_keys_key],
+            args=[])
+
+    @classmethod
+    def switch_profiles(cls,
+                        user_id,
+                        new_profile_key,
+                        old_profile_keys):
+
+        return cls.redis_client.switch_profiles(
+            keys=[
+                cls._profile_pool_key(user_id=user_id),
+                cls._all_profile_keys_key()
+            ],
+            args=[
+                json.dumps(old_profile_keys),
+                new_profile_key
+            ]
+        )
+
+    def store_user_pointers(self):
+
+        pointers = self.redis_client.get_all_user_pointers_if_main_profile(
+            keys=[self.all_pointers_key],
+            args=[self.user_id])
+
+        pointers = [p.decode() if isinstance(p, bytes) else p for p in pointers]
+        pointers = [p[len(self.pointer_prefix()):] for p in pointers]
+
+        max_tries = 3
+        error = None
+        errors = []
+        for pointer in pointers:
+            for i in range(max_tries):
+                try:
+                    self.django_pointer_model.objects.update_or_create(
+                        pointer=pointer,
+                        defaults={'user_id': self.user_id}
+                    )
+                    error = None
+                    break
+                except Exception as e:
+                    if 'OperationalError' in e.__class__.__name__:
+                        error = e
+                        time.sleep(1.0 + random.random() / 10)
+                    else:
+                        raise e
+            if not error is None:
+                errors.append(error)
+
+        if len(errors) > 0:
+            raise errors[0]
+
+        return len(pointers)
+
+    def reset_user_not_found(self):
+
+        self.redis_client.delete(self.user_not_found_in_db_key)
+
+    def remove_from_to_be_cleaned_queue(self):
+
+        self.redis_client.srem(self.users_to_be_cleaned_key, self.user_id)
 
     @classmethod
     def get_stitched(cls,
@@ -296,8 +407,6 @@ class Profile(object):
         # if fallback is true, retrieve from database for those users (if users are missing in db, something is wrong,
         # however we should still delete them from stitch pool)
 
-
-
     def to_cache(self, key, add_to_pool=False, pipe=None):
 
         if pipe:
@@ -355,7 +464,6 @@ class Profile(object):
 
         return result
 
-
     @classmethod
     def from_database(cls,
                       key,
@@ -411,6 +519,45 @@ class Profile(object):
 
         return profile
 
+    @classmethod
+    def get_to_be_cleaned_users(cls,
+                                number=1000):
+        user_ids = cls.redis_client.srandmember(cls._users_to_be_cleaned_key(), number=number)
+        return [user_id.decode() if isinstance(user_id, bytes) else user_id for user_id in user_ids]
+
+    @classmethod
+    def get_expired(cls,
+                    num=300,
+                    start=0,
+                    dev=False):
+
+        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
+        expired_profiles = cls.redis_client.zrangebyscore(cls._all_profile_keys_key(),
+                                                          min='-inf',
+                                                          max='+inf' if dev else (now - cls.profile_ttl),
+                                                          start=start,
+                                                          num=num,
+                                                          withscores=True)
+        return [cls(key=profile_key.decode(),
+                    time=pytz.utc.localize(datetime.datetime.utcfromtimestamp(timestamp)),
+                    init=False) for profile_key, timestamp in expired_profiles]
+
+    def lock(self, ttl):
+        return Lock(self.profile_id_from_key(self.key), self.redis_client, ttl=ttl * 1000)
+
+    def load_and_finalize(self):
+
+        data = self.redis_client.finalize_profile(keys=[self.key], args=[])
+
+        if data is None:
+            raise exceptions.ProfileNotFoundError
+
+        user_hash_key, currently_active, profile_data = data
+        self.user_id = self.user_id_from_user_hash_key(user_hash_key.decode())
+        self.currently_active = currently_active == b'1'
+        profile_data = dict(zip(profile_data[::2], profile_data[1::2]))
+        self.profile_data = {k.decode(): v.decode() if isinstance(v, bytes) else v for k, v in profile_data.items()}
+
     def completely_remove_from_cache(self):
 
         self.redis_client.completely_remove_from_cache(
@@ -508,6 +655,15 @@ class Profile(object):
             return None
         return pytz.utc.localize(datetime.datetime.utcfromtimestamp(float(value)))
 
+    @classmethod
+    def get_user_ids_with_multiple_profiles(cls,
+                                            num,
+                                            start):
+
+        result = cls.django_profile_model.objects.all().values('user_id').annotate(
+            count=cls.django_count_func('user_id')).filter(count__gt=1).order_by()[start:start + num]
+        return [str(res['user_id']) for res in result]
+
     @classmethod
     def multiple_from_database(cls,
                                keys,
@@ -574,6 +730,10 @@ class Profile(object):
     def _user_hash_key(cls, user_id):
         return f'u:h:{cls.domain}:{user_id}'
 
+    @classmethod
+    def user_id_from_user_hash_key(cls, user_hash_key):
+        return user_hash_key.split(':')[-1]
+
     @staticmethod
     def profile_id_from_key(key):
 
@@ -582,7 +742,12 @@ class Profile(object):
     @property
     def all_profile_keys_key(self):
 
-        return f'u:pk:{self.domain}:all'
+        return self._all_profile_keys_key()
+
+    @classmethod
+    def _all_profile_keys_key(cls):
+
+        return f'u:pk:{cls.domain}:all'
 
     @property
     def profile_pool_key(self):
